/**
 * Copyright 2007-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.jbuilding;

import java.io.File;
import java.util.Collection;
import java.util.Map;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import org.ow2.util.maven.jbuilding.jar.ClientJarProvider;

/**
 * Create a big jar archive from the listed bundles content.
 * @since 2.0.0
 */
@Mojo(name = "client-jar",
      defaultPhase = LifecyclePhase.PROCESS_SOURCES)
public class GenerateClientJarMojo extends AbstractResolverMojo {

    /**
     * iPOJO Classifier default name.
     */
    private static final String IPOJO_CLASSIFIER = "ipojo";

    /**
     * Location new resources directory.
     *
     * @since 2.0.0
     */
    @Parameter(defaultValue = "${project.build.directory}/assemble-resources")
    private File output;

    /**
     * Where do I place my artifacts ?
     *
     * @since 2.0.0
     */
    @Parameter(defaultValue = "lib")
    private String directory;

    /**
     * The main class of the Client jar (can be null).
     * @since 2.0.0
     */
    @Parameter
    private String clientJarMainClass;

    /**
     * Generated jar name.
     * @since 2.0.0
     */
    @Parameter(defaultValue = "client.jar")
    private String jarName;

    /**
     * Allow timestamp in SNAPSHOT version if set to true
     * @since 2.0.0
     */
    @Parameter(defaultValue = "false")
    private boolean allowTimestamp = false;

    /**
     * @throws MojoExecutionException When .
     * @see org.ow2.util.maven.jbuilding.AbstractResolverMojo#execute()
     */
    public void execute() throws MojoExecutionException {

        Map<String, Collection<Artifact>> map = resolveArtifacts(CLIENT_SIDE, BOTH_MODE, ALL_ARTIFACT_ITEMS, this.allowTimestamp);

        // OK, now we're sure that everything is in the local repository
        // We have to copy theses files in the output directory
        File libDirectory = new File(output, directory.replace('/',
                                                               File.separatorChar));

        // Create the directory
        if (!libDirectory.exists()) {
            if (!libDirectory.mkdirs()) {
                throw new MojoExecutionException("Cannot create directory: "
                                                 + libDirectory);
            }
        }

        ClientJarProvider provider = new ClientJarProvider();

        File target = new File(libDirectory, jarName);
        provider.setTargetJarFile(target);

        addDefaultExclusionPatterns(provider);

        // List all plans
        for (Collection<Artifact> artifacts: map.values()) {
            // List all artifacts/bundle
            for (Artifact artifact : artifacts) {

                Artifact updated = artifact;

                // If this is an ipojo classified object, switch to the "normal" one
                if (IPOJO_CLASSIFIER.equals(artifact.getClassifier())) {
                    // Change classifier to "normal" one: 'jar'
                    updated = repositorySystem.createArtifact(artifact.getGroupId(),
                            artifact.getArtifactId(),
                            artifact.getVersion(),
                            artifact.getScope(),
                            artifact.getType());
                    try {
                        // Resolve the artifact
                        updated = resolveArtifact(updated);
                    } catch (Exception e) {
                        throw new MojoExecutionException("Cannot resolve artifact: " + updated, e);
                    }
                }

                getLog().debug("Artifact file path: " + updated.getFile());

                // Get the Artifact local File path
                // and add it to the provider sources list
                File source = updated.getFile();
                provider.addSourceFile(source);
            }
        }

        try {
            provider.createClientJarFile(project.getVersion(), clientJarMainClass);
        } catch (Exception e) {
            throw new MojoExecutionException("Cannot create target file: " + target, e);
        }

        project.addCompileSourceRoot(output.getPath());

    }

    /**
     * Assign all the default excluded pattern in the provider.
     * @param provider {@link ClientJarProvider} instance.
     */
    protected void addDefaultExclusionPatterns(final ClientJarProvider provider) {
        // Some excluded patterns (Tomcat, Bouncycastle, some Keys) ...
        provider.addExclusionPattern("org/apache/tomcat");
        provider.addExclusionPattern("org/apache/catalina");
        provider.addExclusionPattern("org/apache/coyote");
        provider.addExclusionPattern("org/bouncycastle");
        provider.addExclusionPattern("org/jgroups");
        provider.addExclusionPattern("org/mortbay");
        provider.addExclusionPattern("META-INF/maven");
        provider.addExclusionPattern("META-INF/BCKEY");
    }

}
