/**
 * Copyright 2007-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.jbuilding.jar;

import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

import org.ow2.util.file.FileUtils;
import org.ow2.util.file.FileUtilsException;

/**
 * Generates a big jar with the bundles content.
 * @author Guillaume Sauthier
 */
public class ClientJarProvider {

    /**
     * Source directories.
     */
    private List<File> watched;

    /**
     * Source directories.
     */
    private List<File> sources;

    /**
     * Target file to create if inexistant or outdated.
     */
    private File target;

    /**
     * Generated jar content.
     */
    private List<String> content;

    /**
     * Excluded patterns.
     */
    private List<String> exclusions;

    /**
     * Default constructor.
     */
    public ClientJarProvider() {
        this.watched = new ArrayList<File>();
        this.sources = new ArrayList<File>();
        this.content = new ArrayList<String>();
        this.exclusions = new ArrayList<String>();
    }

    /**
     * @throws Exception if unable to create the ClassLoader
     */
    public void createClientJarFile(final String version, final String mainClass) throws Exception {

        // Is there a target somewhere ?
        if (!target.exists() || isOutOfDate()) {
            // non existant or out of date
            // re-creates it
            createTargetJarFile(version, mainClass);
        }
    }

    /**
     * Initialize the client jar file.
     * @throws IOException if libraries pack/unpack fails.
     */
    private void createTargetJarFile(final String version, final String mainClass) throws IOException {

        // Delete the existing outdated file if there was one.
        if (target.exists()) {
            target.delete();
        }

        // Build the META_INF/MANIFEST.MF
        content.add("META-INF/MANIFEST.MF");
        Manifest m = new Manifest();
        Attributes attr = m.getMainAttributes();
        attr.put(Attributes.Name.MANIFEST_VERSION, "1.0"); // mandatory!
        attr.put(Attributes.Name.IMPLEMENTATION_VERSION, version); // for getVersion()
        if (mainClass != null) {
            attr.put(Attributes.Name.MAIN_CLASS, mainClass); // set the main class.
        }


        // Create the target file
        OutputStream os = new FileOutputStream(target);
        JarOutputStream jos = new JarOutputStream(os, m);

        // Iterates over the watched directories
        for (File directory : watched) {

            // Only accept file ending with .jar
            File[] childs = directory.listFiles(new FileFilter() {

                public boolean accept(final File pathname) {
                    return pathname.isFile()
                           && pathname.getName().endsWith(".jar")
                           && !pathname.equals(target);
                }

            });

            // Iterates over the bundles
            for (int i = 0; i < childs.length; i++) {
                File bundle = childs[i];

                // Add the bundle content in the client.jar
                processJarFile(bundle, jos);
            }

        }

        // Iterates over the jar sources
        for (File jar : sources) {
            processJarFile(jar, jos);
        }

        // Finish the client.jar
        jos.flush();
        jos.close();

    }

    /**
     * Extract the jar file content in the client.jar.
     * @param bundle Jar file to be processed.
     * @param jos Output
     * @throws IOException if some pack/unpack operations were not run successfully
     */
    @SuppressWarnings("unchecked")
    private void processJarFile(final File bundle, final JarOutputStream jos) throws IOException
    {
        // Look inside of the bundle for all .jar
        JarFile jf = new JarFile(bundle);

        // Pack this jar
        pack(jf, jos);

        // Browse the manifest for ClassPath entries
        Manifest manifest = jf.getManifest();

        String bundleClasspath = manifest.getMainAttributes().getValue("Bundle-Classpath");
        if (bundleClasspath != null) {

            File unpacked = unpack(jf);

            // Got a Bundle: add its inner classpath
            String[] pathElements = bundleClasspath.split(",");

            for (int j = 0; j < pathElements.length; j++) {
                String path = pathElements[j];

                // Do not manage '.'
                if (!".".equals(path)) {

                    File resource = new File(unpacked, path);
                    if (resource.isFile()) {
                        // unpack the jar content.
                        pack(new JarFile(resource), jos);
                    }
                }
            }
        }

        // Do not forget to close
        jf.close();
    }

    private File unpack(final JarFile jarFile) throws IOException {

        try {
            File tmp = FileUtils.createTemporaryDirectory();
            FileUtils.unpack(jarFile, tmp);
            return tmp;
        } catch (FileUtilsException fue) {
            IOException ioe = new IOException("Unable to unpack " + jarFile.getName());
            ioe.initCause(fue);
            throw ioe;
        }
    }

    /**
     * Pack the source content.
     * @param source Jar file.
     * @param jos the output stream
     * @throws IOException if pack fail
     */
    private void pack(final JarFile source, final JarOutputStream jos) throws IOException {

        // Pack all the source entries
        for (Enumeration<JarEntry> e = source.entries(); e.hasMoreElements();) {
            JarEntry entry = e.nextElement();

            // Do not copy jar files, excluded content or already added entries
            if (!entry.getName().endsWith(".jar")
                && !content.contains(entry.getName())
                && !match(entry.getName())) {

                InputStream is = source.getInputStream(entry);

                JarEntry je = new JarEntry(entry.getName());
                jos.putNextEntry(je);

                // Copy the content
                int read = 0;
                final int MAX = 4096;
                byte[] buffer = new byte[MAX];
                while ((read = is.read(buffer)) != -1) {
                    jos.write(buffer, 0, read);
                }

                content.add(entry.getName());
                is.close();
            }
        }
    }

    /**
     * Simple pattern matching method (startsWith).
     * @param name name to be matched against exclusions patterns.
     * @return true if the name match, false otherwise
     */
    private boolean match(final String name) {
        for (String pattern : this.exclusions) {
            boolean match = name.startsWith(pattern);
            if (match) {
                return true;
            }
        }
        return false;
    }

    /**
     * @return true if the target if older than one of its sources.
     */
    private boolean isOutOfDate() {

        final long lastTargetChange = target.lastModified();
        boolean aligned = true;

        // Iterates over the watched directories
        for (File directory : watched) {

            // Only accept file ending with .jar
            File[] newer = directory.listFiles(new FileFilter() {

                public boolean accept(final File pathname) {
                    return pathname.isFile()
                           && pathname.getName().endsWith(".jar")
                           && pathname.lastModified() > lastTargetChange;
                }

            });

            if (newer.length != 0) {
                // newer files
                aligned &= false;
            }
        }

        // Iterates over the sources files
        for (File source : sources) {

            boolean newer = source.getName().endsWith(".jar")
                             && source.lastModified() > lastTargetChange;

            if (newer) {
                // newer files
                aligned &= false;
            }
        }

        return !aligned;
    }

    /**
     * Add a directory whose contained jar files will be processed.
     * @param dir watched directory
     */
    public void addWatchDirectory(final File dir) {
        watched.add(dir);
    }

    /**
     * Add a jar file name to be processed.
     * @param source source jar file name
     */
    public void addSourceFile(final File source) {
        sources.add(source);
    }

    /**
     * Set the target File.
     * @param target target File
     */
    public void setTargetJarFile(final File target) {
        this.target = target;
    }

    /**
     * Add an exclusion pattern.
     * @param pattern exclusion pattern
     */
    public void addExclusionPattern(final String pattern) {
        this.exclusions.add(pattern);
    }

}
